#include <stdio.h>
#include <nicklib.h>

#define NUMBER_OF_EIGENVECTORS_DESIRED 10
#define INPUT_FILE "DATA.INDIVIDUALS"
#define MAXM 10000 /* max number of SNPs */
#define MAXN 500   /* max number of individuals */
#define SCALE_XTX 1
#define VERBOSE 0 /* 1 is X, U, 2 is U and XV and X - U S VT */
#define ROW_MEAN_ADJUST 1
#define ROW_MEAN_ADJUST_NORMALIZE 0
#define ROW_MEAN_ADJUST_NORMALIZE_BOUND 0.01
#define WEIGHT_ROW_MEAN_ADJUST 0
#define COLUMN_ADJUST 0
#define SUBTRACT_FROM_XTX 0
#define VAR_ADJUST 0

int nSNP, NSAMPLES;

int main(void)
{
  int N, x;
  double *Xorig, *X, *weight, weightsum, *COV;
  double *V, *XV, *U, *Svector;
  double *XTX, *evec, *eval; 
  FILE *fp;
  int m, n, nn, rowvalid, nvaleig;
  double rowsum, rowadjust, ans;
  double columnsum, columnadjust;
  double minXTX, sumab, suma, sumb, sum1;
  double normfactor;

  /* malloc */
  if((weight = (double *) malloc(MAXN*sizeof(*weight))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((Xorig = (double *) malloc(MAXM*MAXN*sizeof(*Xorig))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((X = (double *) malloc(MAXM*MAXN*sizeof(*X))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((COV = (double *) malloc(MAXN*MAXN*sizeof(*COV))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((XV = (double *) malloc(MAXM*MAXN*sizeof(*XV))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((U = (double *) malloc(MAXM*MAXN*sizeof(*U))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((Svector = (double *) malloc(MAXN*sizeof(*Svector))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((V = (double *) malloc(MAXN*MAXN*sizeof(*V))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }

  if((XTX = (double *) malloc(MAXN*MAXN*sizeof(*XTX))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((evec = (double *) malloc(MAXN*MAXN*sizeof(*evec))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((eval = (double *) malloc(MAXN*sizeof(*eval))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }

  /* get data */
  if( (fp = fopen(INPUT_FILE, "r")) == NULL)
  {
    fprintf(stderr,"Could not open input file %s\n", INPUT_FILE);  exit(1);
  }
  fscanf(fp,"%d",&nSNP);
  if(nSNP > MAXM) 
  {
    fprintf(stderr,"OOPS nSNP=%d but MAXM is only %d\n",nSNP,MAXM);
    fprintf(stderr,"Please raise MAXM and rerun.\n");
    exit(1);
  }
  fscanf(fp,"%d",&NSAMPLES);
  if(NSAMPLES >MAXN)
  {
    fprintf(stderr,"OOPS NSAMPLES=%d but MAXN is only %d\n",NSAMPLES,MAXN);
    fprintf(stderr,"Please raise MAXN and rerun.\n");
    exit(1);
  }
  if(NSAMPLES < NUMBER_OF_EIGENVECTORS_DESIRED)
  {
    fprintf(stderr,"OOPS NUMBER_OF_EIGENVECTORS_DESIRED=%d but NSAMPLES is only %d\n",NUMBER_OF_EIGENVECTORS_DESIRED,NSAMPLES);
    fprintf(stderr,"Please raise MAXN and rerun.\n");
    exit(1);
  }
  printf("nSNP=%d NSAMPLES=%d NUMBER_OF_EIGENVECTORS_DESIRED=%d\n",nSNP,NSAMPLES,NUMBER_OF_EIGENVECTORS_DESIRED);
  for(m=0; m<nSNP; m++)
  {
    for(n=0; n<NSAMPLES; n++) 
    {
      fscanf(fp,"%d",&N); 
      fscanf(fp,"%d",&x);
      if(N == 0) Xorig[NSAMPLES*m+n] = -100.0;
      else Xorig[NSAMPLES*m+n] = ((double)x)/((double)N);
    }
  } 

  /* mean-adjust each SNP (row) */
  for(m=0; m<nSNP; m++)
  { 
    rowadjust = 0.0;
    rowvalid = 0;
    rowsum = 0.0;
    for(n=0; n<NSAMPLES; n++)
    {
      if(Xorig[NSAMPLES*m+n] >= -99.0)
      {
        rowvalid++;
        rowsum += Xorig[NSAMPLES*m+n];
      }
    }
    if(rowvalid > 0) rowadjust = -rowsum/((double)rowvalid);
    for(n=0; n<NSAMPLES; n++)
    {
      if(Xorig[NSAMPLES*m+n] >= -99.0) 
      {
        X[NSAMPLES*m+n] = Xorig[NSAMPLES*m+n] + rowadjust; 
        if(ROW_MEAN_ADJUST == 0) X[NSAMPLES*m+n] -= rowadjust;
        if(ROW_MEAN_ADJUST_NORMALIZE) 
	{
          normfactor = (-rowadjust)*(2.0+rowadjust);
          if(normfactor < ROW_MEAN_ADJUST_NORMALIZE_BOUND)
            normfactor = ROW_MEAN_ADJUST_NORMALIZE_BOUND;
          X[NSAMPLES*m+n] /= normfactor;
	}
      }
      else 
      {
        X[NSAMPLES*m+n] = 0.0;
        if(ROW_MEAN_ADJUST == 0) X[NSAMPLES*m+n] -= rowadjust;
        if(ROW_MEAN_ADJUST_NORMALIZE) 
	{
          normfactor = (-rowadjust)*(2.0+rowadjust);
          if(normfactor < ROW_MEAN_ADJUST_NORMALIZE_BOUND)
            normfactor = ROW_MEAN_ADJUST_NORMALIZE_BOUND;
          X[NSAMPLES*m+n] /= normfactor;
	}
      }
    }
  }  
  if(WEIGHT_ROW_MEAN_ADJUST)
  {
    /* compute weights from empirical Cov */
    for(n=0; n<NSAMPLES; n++)
    {
      weight[n] = 0.0;
      for(nn=0; nn<NSAMPLES; nn++)
      {
	sumab = 0.0; suma = 0.0; sumb = 0.0; sum1 = 0.0;
	for(m=0; m<nSNP; m++)
	{
	  if((Xorig[NSAMPLES*m+n] < -99.0) || (Xorig[NSAMPLES*m+nn] < -99.0)) continue;
	  sumab += Xorig[NSAMPLES*m+n] * Xorig[NSAMPLES*m+nn];
	  suma += Xorig[NSAMPLES*m+n];
	  sumb += Xorig[NSAMPLES*m+nn];
	  sum1 += 1.0;
	}
	COV[NSAMPLES*n+nn] = (sumab/sum1) - (suma/sum1)*(sumb/sum1);
        weight[n] += COV[NSAMPLES*n+nn];
      }
    }
    /* weight[0] = 1.0/1.44; weight[1] = 1.0/2.43; weight[2] = 1.0/1.98; */
    printf("Weights are %.03f %.03f %.03f\n",weight[0],weight[1],weight[2]);

    for(m=0; m<nSNP; m++)
    { 
      weightsum = 0.0;
      rowsum = 0.0;
      for(n=0; n<NSAMPLES; n++)
      {
        if(Xorig[NSAMPLES*m+n] >= -99.0)
        {
          weightsum += weight[n];
          rowsum += weight[n]*Xorig[NSAMPLES*m+n];
        }
      }
      if(weightsum > 0.0) rowadjust = -rowsum/weightsum;
      for(n=0; n<NSAMPLES; n++)
      {
        if(Xorig[NSAMPLES*m+n] >= -99.0) 
          X[NSAMPLES*m+n] = Xorig[NSAMPLES*m+n] + rowadjust; 
        else X[NSAMPLES*m+n] = 0.0;
      }
    }  
  }
  if(COLUMN_ADJUST)
  for(n=0; n<NSAMPLES; n++)
  { 
    columnsum = 0.0;
    for(m=0; m<nSNP; m++) columnsum += X[NSAMPLES*m+n];
    for(m=0; m<nSNP; m++)
      X[NSAMPLES*m+n] -= columnsum / ((double)nSNP); 
  }  
  if(VAR_ADJUST)
  {
    for(m=0; m<nSNP; m++)
    { 
      rowsum = 0.0;
      for(n=0; n<NSAMPLES; n++) rowsum += X[NSAMPLES*m+n] * X[NSAMPLES*m+n];
      if(rowsum > 0.0) 
      {
	for(n=0; n<NSAMPLES; n++) X[NSAMPLES*m+n] /= sqrt(rowsum);
      }
    }  
  }

  /* compute XTX */
  minXTX = -1.0;
  for(n=0; n<NSAMPLES; n++)
  {
    for(nn=0; nn<NSAMPLES; nn++) 
    {  
      XTX[NSAMPLES*n+nn] = 0.0;
      for(m=0; m<nSNP; m++)
        XTX[NSAMPLES*n+nn] += X[NSAMPLES*m+n]*X[NSAMPLES*m+nn];
      if(SCALE_XTX > 0) XTX[NSAMPLES*n+nn] /= ((double)nSNP);
      if((minXTX<0) || (XTX[NSAMPLES*n+nn] < minXTX)) minXTX = XTX[NSAMPLES*n+nn];
    }
  }
  if(SUBTRACT_FROM_XTX)
  {
    for(n=0; n<NSAMPLES; n++)
    {
      for(nn=0; nn<NSAMPLES; nn++) XTX[NSAMPLES*n+nn] -= minXTX;
    }
  }

  /* EIGXSYM */
  eigxsym(evec, eval, XTX, &nvaleig, NSAMPLES);

  /* further manual calculations */
  for(n=0; n<NSAMPLES; n++)
  {
    if(eval[NSAMPLES-1-n] >= 0.0) Svector[n] = sqrt(eval[NSAMPLES-1-n]);
    else Svector[n] = 0.0;
  }
  printf("\n");
  printf("First %d eigenvalues of variance matrix:\n",NUMBER_OF_EIGENVECTORS_DESIRED);
  printf("(to get singular values of SVD, take sqrt of each eigenvalue)\n");
  for(n=0; n<NUMBER_OF_EIGENVECTORS_DESIRED; n++) printf("  %.02f",Svector[n]*Svector[n]);
  printf("\n\n");

  fprintf(stderr,"First %d eigenvalues of variance matrix:\n",NUMBER_OF_EIGENVECTORS_DESIRED);
  fprintf(stderr,"(to get singular values of SVD, take sqrt of each eigenvalue)\n");
  for(n=0; n<NUMBER_OF_EIGENVECTORS_DESIRED; n++) fprintf(stderr,"  %.02f",Svector[n]*Svector[n]);
  fprintf(stderr,"\n");

  for(n=0; n<NSAMPLES; n++)
  {
    for(nn=0; nn<NSAMPLES; nn++)
      V[n*NSAMPLES+nn] = evec[(NSAMPLES-1-nn)*NSAMPLES+n];
  }
  printf("First %d columns of Matrix V where X = V D Vtranspose\n",NUMBER_OF_EIGENVECTORS_DESIRED);
  printf("  for diagonal matrix D of eigenvalues\n");
  printf("Column k of V represents eigenvector k (with kth eigenvalue above)\n");
  printf("Note that the relative size of the eigenvalues indicates the\n");
  printf("proportion of overall variance described by that eigenvector:\n");
  for(n=0; n<NSAMPLES; n++)
  {
    for(nn=0; nn<NUMBER_OF_EIGENVECTORS_DESIRED; nn++) 
    {
      printf(" ");
      if(V[NSAMPLES*n+nn] >= 0) printf("+");
      printf("%.02f",V[NSAMPLES*n+nn]);
    }
    printf("\n");
  }
}


