#include  <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>  

#include <strsubs.h> 
#include <vsubs.h> 
#include <ranmath.h>  
#include "emvals.h" 
#include "emsubs.h"  
#include "heapsubs.h" 

#define ENDT  -9999

static int debug = NO ;
static int vmode = NO ;
static int wmsg = NO ;

static double **pparams ;
static double **gammas  ;
static int xlen, ppow, xwin ;
static    int *badtime ;
static    int *nnext ;  
static    int ***lnext ;  

static    int *nlist ;
static    int ***pplist ;

static FILE *cfile ;
static int numout ;
static int **decode ;

/** 
 symbol to print in branch 
*/


void cleargams() ;
void settabs(int *pa, int *pb, int n)  ;
void freetabs() ;
void gentabs(int **transl, int *ntransl, 
 int **ltl, int nltl, int ** rtl, int nrtl)  ;

void emglist(int *la, int *lb, int *nla, int *nlb, 
 int **xtab , int bst, int bend) ;
void printxtab(int **xtab, int n) ;

void genl1(int a, int b, int *ll, int *va, int *vb)  ;
double dogamma(int *rdata, int m,int  n)  ;
double dog1(int *pa, int *pb,int n) ;
double balxx(double *a, int n) ;

void doreest() ;

void dodynam(int *pdata, int *rdata,int m,int  n)  ;
void dod1(int *pa, int *pb,int n) ;

void pbranch(int *buff, int xtime, int blen, 
   double prob, double thresh, int state) ;

int msk(int n)  ;
int right (int k, int n) ;
int left (int k, int n)  ;


void setsparam(int tmaxlist, int tmaxn, double tpriorfac, double tpriorg) 
{

 if (tmaxlist>0) smaxlist = tmaxlist ;
 if (tmaxn>0) smaxn = tmaxn ;
 if (tpriorfac>=0.0) spriorfac = tpriorfac ; 
 if (tpriorg>=0.0) spriorg = tpriorg ; 

 printf("## smaxlist: %d\n",smaxlist) ;
 printf("## smaxn: %d\n",smaxn) ;
 printf("## spriorfac: %9.3f\n",spriorfac) ;
 printf("## spriorg: %9.3f\n",spriorg) ;

}

void hmmphase(int *pdata, int *rdata, int m, int n) 
{

   int *ppdata,  *rrdata ;
   int na, nh, lo, hi, hkeep ;  
   int stm, stp, k, x, i ;
   int iter ;
   double ylike, ybase, y, oldylike ;
   int *pa, *pb ;

   xwin = stp = MIN(n, smaxn) ;
   stm = stp -1 ;
   ppow = (int) pow(2.0, (double) stp) ;
   xlen = n-stp+1 ;
   ZALLOC(pparams, xlen, double *) ;
   ZALLOC(gammas, xlen, double *) ;
   ZALLOC(badtime, xlen, int) ;
   ZALLOC(nnext, xlen, int) ;
   for (k=0; k<xlen; k++) { 
    ZALLOC(pparams[k], ppow, double) ;
    ZALLOC(gammas[k], ppow, double) ;
    for (i=0; i<ppow; i++) { 
     gammas[k][i] = DRAND() + 100.0 ;
    }
   }

   doreest() ;

   for (iter=1; iter<=20 ; ++iter)  {  
    ylike = dogamma(rdata, m, n) ;
    doreest() ;
    
#define YTHRESH 0.1 
    if (iter==1) ybase = ylike ;
    ylike -= ybase ;
    y = pparams[0][0] ;
    printf("iter: %2d like: %9.3f ppar: %12.6f\n",iter,ylike,y) ;
    if ((iter>1) && ((ylike-oldylike) < YTHRESH)) break ;
    oldylike = ylike ;
   }
   dodynam(pdata, rdata, m, n) ;
}
void hmmfree() 
{
   int k ;
   for (k=0; k<xlen; k++) { 
    free(pparams[k]) ;
    free(gammas[k]) ;
   }
   free(pparams) ;
   free(gammas) ;
   free(badtime) ;
   free(nnext) ;
}

void printprobs(char *cname, int n, 
 double thresh, int *cmax, int *cmin) 
/** 
 does a branch and bound through the tree 
 of possible haplos.  
 Work is recursive routine (pbranch)
 cmax, cmin must be passed in to output correct coding 
*/
{
   int *buff, state, i ; 
   double prob ;
   double tthresh = thresh ;

   ZALLOC(buff, n , int) ;
   ZALLOC(decode, n, int *) ;
   for (i=0; i<n; i++) {  
    ZALLOC(decode[i],2,int) ;
    decode[i][0] = cmax[i] ;
    decode[i][1] = cmin[i] ;
   }
   printf("cname: %s\n",cname) ;
   openit(cname, &cfile, "w") ;

   for (;;)  {

    state = 0 ;

    prob = 1.0 ;
    numout = 0 ;

    pbranch(buff, 0, 0, prob, tthresh, state) ;
    printf("numout: %d thresh: %12.6f\n",numout, tthresh) ;

    if (numout>0) break ;
    tthresh /= 10.0 ;

   }

   fclose(cfile) ;
   free(buff) ;

   for (i=0; i<n; i++) {  
    free(decode[i]) ;
   }
   free(decode) ;
}
void pbranch(int *buff, int xtime, int blen, 
   double prob, double thresh, int state)
/** 
 recursively branch through tree of possible haplos 
 terminating if prob < thresh 
*/
{
   int i, nextstate, k, sym, rstate ;
   double nextprob ;

   if (prob<thresh) return ;

   if (debug) {
    printf("zz %d %d %d %12.6f\n",xtime, blen, state, prob) ;
    printi1(buff, blen) ;
   }
   
   if (xtime>=xlen) {  
    fprintf(cfile,"%15.6f ", prob) ;
    for (i=0; i<blen; i++) {
     k = buff[i] ;
     sym = decode[i][k] ;
     fprintf(cfile,"%d",sym) ;
    }
    fprintf(cfile,"\n") ;
    ++numout ;
    return ;
   }
   if (xtime==0) {  
    for (i=0; i<ppow ; i++)  { 
     uncodeit2(buff, xwin, i) ;
     pbranch(buff, 1, xwin, pparams[0][i], thresh, i) ;
    }
    return ;
   }
   for (i=0; i<=1; i++) { 
     buff[blen] = i ;
     rstate = right(state, xwin) ;
     nextstate = (rstate << 1) | i ;
     nextprob = prob*pparams[xtime][nextstate] ;
     pbranch(buff, xtime+1, blen+1, 
      nextprob, thresh, nextstate) ;
   }
}

void emphase(int *pdata, int *rdata, int m, int n) 
/** 
 this is fallback code to deal with large amounts of missing data.  
 should usually not be needed 
*/
{

   int *ppdata,  *rrdata ;
   int na, nh, lo, hi, hkeep ;  
   
   if (n<=smaxn) {  
     emphaseit(pdata, rdata, m, n, NULL) ;
     return ;
   }
   na = smaxn ; nh = na/2 ;
   copyiarr(rdata, pdata, m*n) ;

   ZALLOC(ppdata, m*na, int) ;
   ZALLOC(rrdata, m*na, int) ;
   for (hi = na-1 ; hi < n; hi++)  {
    lo = hi-(na-1) ;
    copyrec(pdata, rrdata, m, n, na, lo, hi, 0) ;
    emphaseit(ppdata, rrdata, m, na, NULL) ;
    if (hi < (n-1)) {
     hkeep = nh ;
    }
    else {
     hkeep = na-1 ;
    }
     copyrec(ppdata, pdata, m, na, n, 0, hkeep, lo) ;
   }
   free(ppdata) ;
   free(rrdata) ;
}

double dogamma(int *rdata,int m,int  n)  
/* gamma pass*/
{
    int *pa, *pb ;
    int j, k ;
    double ylike = 0.0 ;
    cleargams() ;
    for (j=0; j<m/2 ; ++j) {
     k = 2*j ;
     pa = rdata+k*n ;
     pb = pa + n ;
/**
     vmode = NO ;
     if (j<=1) vmode = YES ;
*/
     ylike += dog1(pa,pb,n) ;  
    }
    return ylike ;
}

double dog1(int *pa, int *pb,int n) 
/* gamma 1 individual */
{
     double **alpha ;
     double *afac ;
     double **beta ;
     double *bfac ;
     double ysc, ytmp, y ;
     int w, x, nw, nx, i,j, a0, a1, nnex, s, t ;

     settabs(pa, pb, n) ; 
     ZALLOC(alpha, xlen, double *) ;
     ZALLOC(afac, xlen, double) ;
     ZALLOC(beta, xlen, double *) ;
     ZALLOC(bfac, xlen, double) ;

     for (x=0; x<xlen; x++) {  
      if (x>0) afac[x] = afac[x-1] ;
      if (badtime[x]==1)  continue ;
      nx = nlist[x] ;

      ZALLOC(alpha[x], nx, double) ;
      ZALLOC(beta[x], nx, double) ;
      vclear(beta[x], 1.0, nx) ;

      w = MAX(x-1,0) ;
      if ( (x==0) ||  (badtime[w]>0)) {  
       y = 0.0 ;
       for (i=0; i<nx; i++) {  
        if (pplist[x][i] == NULL) { 
         fatalx("bad bug %d %d %x %x\n",x,i,pplist[x],pplist[x][i]) ;
        }
        a0 = pplist[x][i][0] ;
        a1 = pplist[x][i][1] ;
        alpha[x][i]  = pparams[x][a0]*pparams[x][a1] ;
        if ((vmode)  && (x==0)) {  
         printf("vm: %d %d %d %12.6f %12.6f\n",
          i,a0,a1,pparams[x][a0],alpha[x][i]) ;
        }
       }
       y = asum(alpha[x], nx) ;
       if (y==0.0) { 
        badtime[w] = 2 ;
        continue ;
       }
       y =balxx(alpha[x], nx)  ;
       if (x==0) afac[x] = y ;  
       else afac[x] = afac[x-1] + y ;
       continue ;
      }

      nw = nlist[w] ;
      nnex = nnext[w] ;
      for (i=0; i<nw; i++) {  
       ysc = alpha[w][i] ;
        
       for (t=0; t<nnex; ++t) {
        if (lnext[w][i] == NULL) fatalx("bad lnext %d %d\n",w,i) ;
        j = lnext[w][i][t] ;
        if (j== ENDT) break ;
        if (j>=nx) {  
         fatalx("(dog1) logic bug\n") ;
        }
        a0 = pplist[x][j][0] ;
        a1 = pplist[x][j][1] ;
        y = pparams[x][a0]*pparams[x][a1] ;
        alpha[x][j] += ysc*y ;
       }
      }
       y = asum(alpha[x], nx) ;
       if (y==0.0) { 
        badtime[w] = 2 ;
        continue ;
       }
       y = balxx(alpha[x], nx)  ;
       afac[x] = afac[x-1] + y ;
/* end of forwards pass */
     }
     for (x=xlen-1; x>=1 ; x--) {  
      w = x-1  ;
      bfac[w] = bfac[x] ;

      if (badtime[x]==1) continue ;
      if (badtime[w]>0) continue ;

      nx = nlist[x] ;
      nw = nlist[w] ;
      nnex = nnext[w] ;
      for (i=0; i<nw; i++) {  
       beta[w][i] = 0.0 ;
       for (t=0; t<nnex; ++t) {
        if (lnext[w][i] == NULL) fatalx("bad lnext %d %d\n",w,i) ;
        j = lnext[w][i][t] ;
        if (j== ENDT) break ;
        if (j>=nx) {  
         fatalx("(dog1) logic bug\n") ;
        }
        a0 = pplist[x][j][0] ;
        a1 = pplist[x][j][1] ;
        ysc = beta[x][j] ;
        y = pparams[x][a0]*pparams[x][a1] ;
        beta[w][i] += ysc*y ;
       }
      }
      y =balxx(beta[w], nw)  ;
      bfac[w] = bfac[x] + y ;
/* end of backwards pass */
     }
     for (x=0; x<xlen; ++x) {
       if (badtime[x] >0)  continue ;
       nx = nlist[x] ;
       vvt(alpha[x], alpha[x], beta[x], nx) ;  /* gammas */
       y = asum(alpha[x],nx) ;
       if (y==0.0) {  
        badtime[x] = 2 ;
        continue ;
       }
       vst(alpha[x], alpha[x], 1.0/y, nx) ;
       ysc = afac[x] + bfac[x] + log(y) ;

       for (j=0; j<nx; j++) {  
        a0 = pplist[x][j][0] ;
        a1 = pplist[x][j][1] ;
        y = alpha[x][j] ;
        gammas[x][a0] += y ;
        gammas[x][a1] += y ;
       }
/**
       printf("check: %d %12.6f\n",x,ysc) ;
*/
       if (vmode==YES) printf("vmode %d %9.3f\n",x, afac[x]) ;
     }
     
     y = afac[xlen-1] ;
     for (x=0; x<xlen; ++x) {
       if (badtime[x] == 1)  continue ;
       free(alpha[x]) ;
       free(beta[x]) ;
     }
     freetabs() ;
     free(alpha) ;
     free(beta) ;
     free(afac) ;
     free(bfac) ;
     return y ;
}
double balxx(double *alpha, int len)   
/* balance and return log(scale) */
{
  double y ; 
  y = asum(alpha,len) ;
  vst(alpha, alpha, 1.0/y, len) ;

  return log(y) ;

}
void dodynam(int *pdata, int *rdata,int m,int  n)  
/* dynamic program to decode */
{
    int *pa, *pb ;
    int j, k ;
    copyiarr(rdata, pdata, m*n) ;
/* default */
    for (j=0; j<m/2 ; ++j) {
     k = 2*j ;
     pa = pdata+k*n ;
     pb = pa + n ;
     dod1(pa,pb,n) ;  
    }
}

void dod1(int *pa, int *pb,int n) 
/**
 dyn. program.  1 individual 
 pa, pb are input and output 
*/
{
     double **score ;
     int **blast ;  
     int *bestseq ;
     double ysc, ytmp, y ;
     int w, x, nw, nx, i,j, a0, a1, nnex, s, t ;
     int *ta, *tb, lo, hi ;

     ZALLOC(ta, n, int) ;
     ZALLOC(tb, n, int) ;
     settabs(pa, pb, n) ; /* copyiarr set this */
     ZALLOC(score, xlen, double *) ;
     ZALLOC(blast, xlen, int *) ;

     for (x=0; x<xlen; x++) {  
      if (badtime[x] == 1)  continue ;
      nx = nlist[x] ;
      ZALLOC(score[x], nx, double) ;
      ZALLOC(blast[x], nx, int) ;

      vclear(score[x], -1.0e20, nx) ;
      ivclear(blast[x], -2, nx) ;

      w = MAX(x-1,0) ;
      if ( (x==0) || (badtime[w] >0)) {  
       for (i=0; i<nx; i++) {  
        if (pplist[x][i] == NULL) { 
         fatalx("bad bug %d %d %x %x\n",x,i,pplist[x],pplist[x][i]) ;
        }
        a0 = pplist[x][i][0] ;
        a1 = pplist[x][i][1] ;
        score[x][i] = ysc = log(pparams[x][a0]*pparams[x][a1]) ;
        blast[x][i] = -1 ;
       }
       continue ;
      }

      nw = nlist[w] ;
      nnex = nnext[w] ;
      for (i=0; i<nw; i++) {  
       ysc = score[w][i] ;
        
       for (t=0; t<nnex; ++t) {
        if (lnext[w][i] == NULL) fatalx("bad lnext %d %d\n",w,i) ;
        j = lnext[w][i][t] ;
        if (j==ENDT) break ;
        if (j>=nx) {  
         fatalx("(dod1) logic bug\n") ;
        }
        a0 = pplist[x][j][0] ;
        a1 = pplist[x][j][1] ;
        y = log(pparams[x][a0]*pparams[x][a1]) ;
        ytmp = ysc + y ;
        if (ytmp>score[x][j]) {
         score[x][j] = ytmp ;
         blast[x][j] = i ;
        }
       }
      }
/* end of forwards pass */
     }

     ZALLOC(bestseq, xlen, int) ;
     ivclear(bestseq, -3, xlen) ;

     for (w=xlen-1; w>=0; --w) {  
      if (badtime[w]>0) continue ;
      nw = nlist[w] ;
      x = w+1 ;
      if (w == (xlen-1) || (badtime[x]) )  {          
       vlmaxmin(score[w],nw, &t, NULL) ;
       bestseq[w] = t ;
       continue ;
      }
      s = bestseq[x] ;
      if (s<0) continue ;
      bestseq[w] = blast[x][s] ;
      if (debug)
       fprintf(stderr, "zz2 %d %d %d\n",x,s,bestseq[w]) ;
     }
/** bestseq is now optimal state sequence */
    if (debug){
       printf("zz3 %d %d %d %d\n",badtime[0],xwin,bestseq[0],bestseq[1]) ;
       printi2(bestseq,20) ;
    }

    ivclear(ta, -4, n) ;
    ivclear(tb, -4, n) ;
    for (x=0; x<xlen; x++) { 
     nx = nlist[x] ;
     s = bestseq[x] ;
     if (s<0) continue ;

     if (s >= nx) 
      fatalx("bestseq out of range %d %d %d\n",x,nx,s) ;

     a0 = pplist[x][s][0] ;
     a1 = pplist[x][s][1] ;
     uncodeit2(ta+x, xwin, a0) ;
     uncodeit2(tb+x, xwin, a1) ;
    }
/** 
 don't use DP to decode when there is badtime nearby */
    for (i=0 ; i<n; i++) {  
     if (ta[i] != -4) continue ;
     lo = MAX(0, i-xwin+1) ; 
     hi = MIN(n-1, i+xwin-1) ;
     for (j=lo; j<=hi ; j++)  {  
      ta[j] = pa[j] ;
      tb[j] = pb[j] ;
     }
    }

    copyiarr(ta, pa, n) ;
    copyiarr(tb, pb, n) ;
 
     for (x=0; x<xlen; x++) {  
      if (badtime[x] == 1)  continue ;
       free(score[x]) ;
       free(blast[x]) ;
     }
     free(bestseq) ;
     freetabs() ;
     free(score) ;
     free(blast) ;
     free(ta) ; 
     free(tb) ; 
}

double score1(int *pa, int n) 
/* 
 return score 
*/
{
  int x, k ;
  double ysc ;

  ysc = 1.0 ;
  for (x=0; x<xlen; ++x) {  
   k = codeit2(pa+x, xwin) ;
   if (k<0) 
    fatalx("(score1) input not all filled\n") ;
   ysc *= pparams[x][k] ;
  }
  return ysc ;
}

void cleargams() 
{
#define EPS 1.0e-10 
   int k ;
   for (k=0; k<xlen; k++) { 
    vclear(gammas[k], EPS,  ppow) ;
   }
}
void
settabs(int *pa, int *pb, int n)  
{

    int x, nx, i, nnex, t, q ;
    int heapsw ;

    ZALLOC(nlist, xlen, int) ;
    ZALLOC(pplist, xlen, int **) ;
    ZALLOC(lnext, xlen, int **) ;

    ivmaxmin(pa, n, &t, NULL) ;
    ivmaxmin(pb, n, &q, NULL) ;

    t = MAX(t,q) ;
    if (t>9) fatalx("(settabs) bad data\n") ;

    ivclear(badtime, NO, xlen) ;
    ivclear(nlist, 0, xlen) ;
    ivclear(nnext, 0, xlen) ;

    for (x=0; x<xlen; x++)  {  
     heapsw = NO ;
     nx = lenplist(pa + x, pb + x, xwin) ;
     if (nx>smaxlist) {  
      if (wmsg == NO) {
       fprintf(stderr, "WARNING lots of missing data or hets\n") ;
       wmsg = YES ;
      }
      nx = smaxlist ;
      heapsw = YES ;
     }
     nlist[x] = nx ;

     ZALLOC(pplist[x], nx, int *) ;
     ZALLOC(lnext[x], nx, int *) ;
     for (i=0; i<nx; i++)  { 
      ZALLOC(pplist[x][i], 2, int) ;
      if (x == (xlen-1))  
       ZALLOC(lnext[x][i], 1, int) ;
/** 
 makes freetabs more uniform 
*/
     }
     if (heapsw==NO)  {
      genpair(pplist[x], nx,  
       pa+x, pb+x, xwin) ;
     }
     else  {
      t = lenplist(pa + x, pb + x, xwin) ;
      genpairh(pplist[x], nx,  
       pa+x, pb+x, xwin, x) ;
     }
    }
    for (x=0; x<xlen-1; ++x) {  
     if (badtime[x]) continue ;
     if (badtime[x+1]) { 
       nx = nlist[x] ;
       for (i=0; i<nx; i++) {
        ZALLOC(lnext[x][i], 1, int) ;
       }
      continue ;
     }
     gentabs(lnext[x], &nnext[x], 
        pplist[x], nlist[x], pplist[x+1], nlist[x+1] ) ;
    }
    if (nnext[x] == 0) badtime[x+1] = YES ;
}

void freetabs() 
{
   int x, i, nx ;
   for (x=0; x<xlen; x++)  {
    if (badtime[x]) continue ;

    nx = nlist[x] ;
    if (nx == 0) continue ;
    for (i=0; i<nx; i++)  { 
     if (lnext[x][i] == NULL) fatalx("lnext bug\n") ;
     free(lnext[x][i]) ;
    }
    free(lnext[x]) ;

    for (i=0; i<nx; i++)  { 
     if (pplist[x][i] == NULL) fatalx("pplist bug\n") ;
     free(pplist[x][i]) ;
    }
    free(pplist[x]) ;
   }
   free(pplist) ;
   free(nlist) ;
   free (lnext) ;
}

void gentabs(int **transl, int *ntransl, 
 int **ltl, int nltl, int ** rtl, int nrtl)  

/** 
 a key routine 
 given a list of left and right legal pairs 
 returns a transition list of legals.  
 ntransl[i] is number of successors of ltl[i] 
 t= transl[i][j] indexes successors  
 meaning ltl[i] -> rtl[t] is legal 
 structure here is number of successors is 
 independent of i unless genpairh called.  
*/

{
   int lena, k, i, j ;
   int **xtab, *xind ;
   int w[2] ;
   int la[16], lb[16], nla, nlb ;
   int numtr=-1, lp, rp, maxtr ;
   int bst, bend ;

   maxtr = -1 ;
   lena = nltl+nrtl ;
   ZALLOC(xtab, lena+1, int *) ;
   for (k=0; k<=lena; k++) {  
    ZALLOC(xtab[k], 3, int) ;
   }
   for (i=0; i<nltl; i++) {  
    xtab[i][0] = right(ltl[i][0], xwin) ;
    xtab[i][1] = right(ltl[i][1], xwin) ;
    xtab[i][2] = - (i+1) ;
   }
   for (i=0; i<nrtl; i++) {  
    j = i+nltl ;
    xtab[j][0] = left(rtl[i][0], xwin) ;
    xtab[j][1] = left(rtl[i][1], xwin) ;
    xtab[j][2] = i+1 ;
   }
   ipsortit(xtab, NULL, lena, 3) ;

   copyiarr(xtab[0], w, 2)  ; 
   ivclear(xtab[lena], -2, 3) ;
/* force difference at end */
   bst =  0 ;
   for (k=1; k<=lena; k++) {  
    if (bst == lena) break ;
    j = compiarr(w, xtab[k], 2) ;
    if (j==0) continue ;
     bend = k-1 ;
     emglist(la, lb, &nla, &nlb, xtab, bst, bend) ;
     if (debug)  {
      fprintf(stderr,"zzd %d %d %d %d\n",
        bst, bend, nltl, nrtl) ; 
      for (j=bst; j<=bend; j++)  { 
       fprintf(stderr,"zz1 %d %d %d",j,nla,nlb) ;
       fprintf(stderr," %d %d %d\n",xtab[j][0],xtab[j][1],xtab[j][2]) ;
      }
     }

     numtr = nlb ;
     maxtr = MAX(maxtr, nlb) ;
     copyiarr(xtab[k], w, 2) ; 
     bst = k ;  
     for (i=0; i<nla; i++) { 
      lp = la[i] ;
      ZALLOC(transl[lp], numtr+1, int) ;
      transl[lp][numtr] = ENDT ;  
/** 
 doubtful style alloc in different routine to free 
*/
      for (j=0; j<nlb; j++) { 
       rp = lb[j] ;
       transl[lp][j] = rp ;
      }
     }
   }





   for (k=0; k<lena; k++) {  
    free(xtab[k]) ;
   }
   free(xtab) ;

   *ntransl = maxtr ;

}
void emglist(int *la, int *lb, int *nla, int *nlb, 
 int **xtab , int bst, int bend) 
{  
  int ta, tb, i, t ;
  if ((bst<0) || (bst>bend)) {
   *nla = *nlb = 0; return ;
  }
  ta = tb = 0; 
  for (i=bst ; i<= bend; i++) { 
   if ((ta==16) || (tb==16)) {  
    printxtab(xtab+bst, bend-bst+1) ;
    fflush(stdout) ;
    fatalx("emglist error %d %d\n", bst, bend) ;
   }
   t = xtab[i][2] ;
   if (t<0) {  
    t = -t ; --t ;
    la[ta] = t ;  ++ta ;
    continue ;
   }
   --t ;
   lb[tb] = t ;  ++tb ;
  }
  *nla = ta ;
  *nlb = tb ;
  return ;
}
void printxtab(int **xtab, int n) 
{
  int i ;  
  for (i=0; i<n; i++) {  
    printf("xtab: %5d %10d",i, xtab[i][0]) ;
    printf(" %10d",xtab[i][1]) ;
    printf(" %10d\n",xtab[i][2]) ;
  }
}

void doreest() 

{
  int k, i ;
  double y ;
  k = 0 ;
  y = asum(gammas[k], ppow) ;
  vst(pparams[k], gammas[k], 1.0/y, ppow) ;
  for (k=1; k<xlen; k++) {
   for (i=0; i<ppow; i += 2)  {
    y = asum(gammas[k]+i, 2)  ;
    vst(pparams[k]+i, gammas[k]+i, 1.0/y, 2) ;
   }

  }
  return ;
}

void
emphaseit(int *pdata, int *rdata, int m, int n, double *probs) 

{
  
  int ***pplist , *nlist, maxlist  ;
  int x, i, npow, k, t, l ; 
  double y, *newprobs ;
  double ylike, ybase ;
  int iter ;
  int nx ;
  int nmax, nmin ;  /* max min uncertainty */  
  int *tdata ;
  int *t2data ;
  double *pprobs ;
  

  maxlist = smaxlist ;

  ZALLOC(pplist, m/2, int **) ;
  ZALLOC(nlist, m/2, int)  ;

  setlenlist(nlist, rdata, m, n, n) ;
  for (x=0; x<(m/2); x++)  { 
   nx = nlist[x] ;
   if (debug) 
    printf ("zz %d %d %d\n",x,m/2,nx) ;
   fflush(stdout) ;
   if (nx>maxlist) { 
    nlist[x] = 0 ;
    nx = 0 ;
   }
   if (nx==0) continue ;
   ZALLOC(pplist[x], nx, int *) ;
    for (i=0; i<nx ; i++) { 
     ZALLOC(pplist[x][i], 2, int) ;
    }
  }


   for (x=0; x<m/2 ; ++x) {
    k = 2*x ;
    if (sdebug)
     printf("x: %3d nlist: %5d\n",x,nlist[x]) ;
    nx = nlist[x] ;
    if (nx==0) continue ;
    genpair(pplist[x], nx,  
     rdata+k*n, rdata+(k+1)*n, n) ;
   }

  npow = (int) pow(2.0, (double) n) ;

  ZALLOC(pprobs, npow, double) ;
  ZALLOC(newprobs, npow, double) ;

  for (k=0; k<npow; k++) { 
   pprobs[k] = DRAND() + 100.0 ;
  }
  y = asum(pprobs, npow) ;
  vst(pprobs, pprobs, 1.0/y, npow) ;

  
   for (iter= 1; iter<20; ++iter) {  
    ylike = doemph(pplist, nlist, m, n, pprobs, newprobs) ;
    if (newprobs[0] < 0.0) { 
     printf("(doemph) bug\n") ;
     printimat(rdata, m, n) ;
     fatalx("j0j1 bug\n") ;
    }
    if (iter==1) ybase = ylike ;
    ylike -= ybase ;
    copyarr(newprobs, pprobs, npow) ;
    if (sdebug)
     printf("iter: %d %9.3f\n",iter, ylike) ;
   }


   copyiarr(rdata, pdata, m*n) ;
   emdecode(pdata, pplist, nlist , m, n, pprobs) ;

   if (probs!=NULL) copyarr(pprobs, probs, npow) ;

   free(newprobs) ;

  for (x=0; x<m/2; x++)  { 
    nx = nlist[x] ;
    if (nx==0) continue ;
    for (i=0; i<nx ; i++) { 
     free(pplist[x][i]) ;
    }
    free(pplist[x]) ;
  }

  free(pplist) ;

  
  setlenlist(nlist, pdata, m, n, n) ;
  ivmaxmin(nlist, m/2, &nmax, &nmin) ;
  if ((nmin==0) || (nmax>maxlist)) {  
/* problem.  */
   ZALLOC(tdata, m*n, int) ;
   ZALLOC(t2data, m*n, int) ;
   for (l=n-1; l>=1; l--) {
    setlenlist(nlist, pdata, m, n, l) ;
    ivmaxmin(nlist, m/2, &nmax, &nmin) ;
    if (nmax<=maxlist) break ;
   }
/** now rephase on first l */
    copyrec(pdata, tdata, m, n, l, 0, l-1, 0) ;
    emphaseit(t2data, tdata,  m, l, pprobs)  ;
    copyiarr(pdata, tdata, m*n) ; 
    copyrec(t2data, tdata, m, l, n, 0, l-1, 0)  ;
    emphaseit(pdata, tdata,  m, n, pprobs)  ;
    free(tdata) ;
    free(t2data) ;
   }


  free(pprobs) ;
  free(nlist) ; 

  if (sdebug == NO) return ;
   printf("\n") ;
   for (k=0; k<m; k++) {
    printf("%5d ",k) ;
    printi2(rdata+k*n,n) ;
    printf("%5s ","") ;
    printi2(pdata+k*n,n) ;
    printf("\n") ;
   }
}  
int
copyrec(int *tdata, int *pdata, int m, 
 int tr, int pr, int tlo, int thi, int plo)

{
   int i, j, k1, k2 ;
   for (i=0; i<m; i++) {  
    for (j=tlo; j<=thi ; j++)  {  
     k1 = i*tr + j ;  
     k2 = i*pr + (j-tlo)+ plo ;
     pdata[k2] = tdata[k1] ;
    }
  }  
}

void
emdecode(int *phasedata, int ***pplist, int *nlist , 
  int m, int n, double *pprobs)  

/** 
 em decoding.  most probable answer 
*/

{
 
       int npow, k, kmax, nl, x ;
       int j0, j1 ;
       double y, ymax, xmax ;

       npow = (int) pow(2.0, (double) n) ;
       xmax = m/2 ; --xmax ;
       for (x=0; x<=xmax ; x++)  {  
        nl = nlist[x] ;
        if (nl==0) continue ;  /* no decode */
        ymax = 0.0 ;
        for (k=0; k<nl; k++) {
         j0 = pplist[x][k][0] ;
         j1 = pplist[x][k][1] ;
         if ((j0>=npow) || (j1>=npow)) { 
          printf("j0bug\n") ;
          printimat(phasedata, m, n) ;
          fatalx("range bug\n") ;
         }
         y = pprobs[j0]*pprobs[j1] ;

         if (y>ymax) {  
          kmax = k ;
          ymax = y ;
         }
        }
        k = kmax ;
        j0 = pplist[x][k][0] ;
        j1 = pplist[x][k][1] ;
        uncodeit2(phasedata+(2*x)*n, n, j0) ;
        uncodeit2(phasedata+(2*x+1)*n, n, j1) ;
    }
}  

double
doemph(int ***pplist, int *nlist, int m, int n, 
  double *pprobs, double *newprobs) 

{

       int npow ;
       double *gamma ;
       int nl , k, x, i ;
       int j0, j1 ;
       double y, yp, ylike, xpmax, ybase ;
       double *wp, *wtrans, *wptrans, *wr ;
       int lmax = -1 ;
       static int first = YES ;

       npow = (int) pow(2.0, (double) n) ;
       ZALLOC(gamma, npow, double) ;

       ylike = 0.0 ;

       if (spriorfac>0.0)   {  

        if ((spriorg<=0.0) || (spriorg>=1.0)) {
         fatalx("bad spriorg %9.3f\n",spriorg) ;
        }

/** 
 fast fourier convolution
*/
        ZALLOC(wp, npow, double) ;
        ZALLOC(wtrans, npow, double) ;
        ZALLOC(wptrans, npow, double) ;
        ZALLOC(wr, npow, double) ;
        loadwp(wp, n, spriorg) ;

        fasth(wtrans, wp, n) ;
        fasth(wptrans, pprobs, n) ; 
        vvt(wr, wptrans, wtrans, npow) ;
        fasth(wr, wr, n) ;  /* inverse transform */

        y = asum(wr, npow) ;
        if (y==0.0) fatalx("zero y\n") ;

        y = (double) npow ;
        vst(wr, wr, 1.0/y, npow) ;
        vvt(gamma, wr, pprobs, npow) ;
        yp = asum(gamma, npow) ;

        ylike = spriorfac*log(yp) ;                  
        if (sdebug) 
         printf("initfac %12.6f\n",ylike) ;
        vst(gamma, gamma, spriorfac*2.0/yp, npow) ;
        if (sdebug) {
         vlmaxmin(gamma, npow, &lmax, NULL) ;
         xpmax = gamma[lmax] ;
         printf("lmax:  %d  xpmax: %9.3f\n",lmax, xpmax) ;
        }
        /*END spriorfac > 0 */
       }

       for (x=0; x<m/2 ; x++)  {  
        nl = nlist[x] ;
        if (nl==0) continue ;
        yp = 0.0 ;
        for (k=0; k<nl; k++) {
         j0 = pplist[x][k][0] ;
         j1 = pplist[x][k][1] ;
         if ((j0>=npow) || (j1>=npow)) { 
          printf("j1bug %d %d %d %d %d\n", x,k, npow, j0, j1) ;
          newprobs[0] = -999.0 ;
          return ;
// calling program will abort
         }
         y = pprobs[j0]*pprobs[j1] ;
         yp += y ;
        }
        ylike += log(yp) ; 
        for (k=0; k<nl; k++) {
         j0 = pplist[x][k][0] ;
         j1 = pplist[x][k][1] ;
         if ((j0>=npow) || (j1>=npow)) { 
          printf("xxbug\n %d %d %d %d %d",x,k,npow, j0, j1) ;
          newprobs[0] = -9999.0 ;
          return ;
         }
         y = pprobs[j0]*pprobs[j1]/yp ;
         gamma[j0] += y ;
         gamma[j1] += y ;
        }
       }
       if (lmax>=0) { 
        printf("gamma[lmax]: %9.3f\n",gamma[lmax]) ;
       }
       y = asum(gamma, npow) ;
       vst(newprobs, gamma, 1.0/y, npow) ;
       free(gamma) ;
       if (spriorfac>0.0)   {  
        free(wp) ;
        free(wtrans) ;
        free(wptrans) ;
        free(wr) ;
       }
       return ylike ;
}

void
loadwp(double *wp, int n, double g) {  

  int npowh ;
  double y, gg ;

  gg = MAX(g, .001) ;

  if (n==0) { 
   wp[0] = 1.0 ;
   return ;
  }
  
  loadwp(wp, n-1, gg) ;
  npowh = (int) pow(2.0, (double) (n-1)) ;
  vst(wp+npowh, wp, gg, npowh) ;
  vst(wp, wp, 1.0-gg, npowh) ;

/** 
 simple binomial model 
*/
  wp[0] = 0.0 ;
  y = asum(wp, 2*npowh) ;
  vst(wp, wp, 1.0/y, 2*npowh) ; 
}
void
fasth(double *wtrans, double *wp, int n)  
{
/** 
 fast Hadamard transform 
 written so as to work in-place
*/
  int npowh, i, j ;
  double x, y ;

  if (n==1) {  
   x=wp[0]+wp[1] ;
   y=wp[0]-wp[1] ;
   wtrans[0] = x ;
   wtrans[1] = y ;
   return ;
  }
  npowh = (int) pow(2.0, (double) (n-1)) ;
  fasth(wtrans, wp, n-1) ;
  fasth(wtrans+npowh, wp+npowh, n-1) ;
  for (i=0; i<npowh; i++)  {
   j = i+npowh ;
   x=wtrans[i]+wtrans[j] ;
   y=wtrans[i]-wtrans[j] ;
   wtrans[i] = x ;
   wtrans[j] = y ;
  }
}

void printi2(int *a, int n) 
{
  int j ;
  for (j=0; j<n; j++) { 
   printf("%2d",a[j]) ;
  }
  printf ("\n") ;
}

void
setlenlist(int *nlist, int *rdata, int m, int n, int rlim)  
{
/** 
 set up nlist
 but only look rlim out 
*/ 
  int x, k, nx ;
  for (x=0; x<m/2; x++)  { 
   k = 2*x ;
   nx = lenplist(rdata+k*n, rdata+(k+1)*n, rlim) ;
   nlist[x] = nx ; 
  }
}

int lenplist (int *ia, int *ib, int n) 
/** 
 return number of matching haplo pairs 
*/
{
  int i, t, nx = 1 ;

  for (i=0; i<n; i++) {
   if ((ia[i]>9) || (ib[i]>9)) 
    fatalx("lenpl %d %d %d\n",i,ia[i],ib[i]);
   t = -1; 
   if ((ia[i] == 9) && (ib[i] == 9) && (t<0)) { 
    t=1 ; nx *=2 ;
   }
   if ((ia[i] >= 0) && (ib[i] >= 0) && (t<0)) { 
    t=0 ;
   }
   if ((ia[i] < 0) && (ib[i] < 0) && (t<0)) { 
    t=4 ; nx *=4 ;
   }
   if ((ia[i] >= 0) && (ib[i] < 0) && (t<0)) { 
    t=2 ; nx *= 2 ;
   }
   if ((ia[i] <0)  && (ib[i] >= 0) && (t<0)) { 
    t=3 ;  nx *= 2 ;
   }
   if (t<0) fatalx("bug %d %d\n",ia[i],ib[i]) ;
  }
  return nx ;
}

void 
genpairh(int **pplist, int nlist, 
  int *ia, int *ib, int n, int x)  
/** 
 Like genpair but HEAP version
*/
{
  int i, j, k, l, t0, t1 ;
  int **wa, **xa ;
  int n1, n2, l1, l2, mul ;
  HEAPX **hh, *pq, htemp ;
  double ysc, y ;

  if (n==0) return ;
  ZALLOC(hh, nlist, HEAPX *) ;
  for (i=0; i<nlist; i++)  {
    hh[i] = (HEAPX *) malloc(sizeof(HEAPX)) ;
  }
  pq = &htemp ;

  if (n==1)  {  
   fatalx("genpairh called on just one base\n") ;
  }

  n1 = n/2 ;
  n2 = n-n1 ;
  l1 = lenplist(ia, ib, n1) ;
  l2 = lenplist(ia+n1, ib+n1, n2) ;

  if ((l1*l2) <= nlist) {
   printi1(ia,n) ;
   printi1(ib,n) ;
   fatalx("(genpairh) logic bug\n") ;
  }

  ZALLOC(wa, l1, int *) ;
  for (i=0; i<l1; i++) {  
    ZALLOC(wa[i], 2, int) ;
  }

  ZALLOC(xa, l2, int *) ;
  for (i=0; i<l2; i++) {  
    ZALLOC(xa[i], 2, int) ;
  }

  genpair(wa, l1, ia, ib, n1) ;
  genpair(xa, l2, ia+n1, ib+n1, n2) ;

  mul = (int) pow(2.0, (double) n2) ;
  clearheap(hh, nlist) ;
  for (i=0; i<l1; i++) {
   for (j=0; j<l2; j++) {
    t0 = mul*wa[i][0]+xa[j][0] ;
    t1 = mul*wa[i][1]+xa[j][1] ;
    y = pparams[x][t0]*pparams[x][t1] + 1.0e-6 ;
    ysc = log(y) ;
    if (isnan(ysc))  {  
     printf("%d %d %d %g %g %g\n",
      x,t0,t1,pparams[x][t0],pparams[x][t1], y) ;
     fatalx("bad ysc\n") ;
    }

    if (ysc < hh[0]->score) continue ;
    pq->a1=t0 ;
    pq->a2=t1 ;
    pq->score = ysc ;
    add1heap(hh, nlist, pq) ;
   }
  }

  for (k=0; k<nlist; k++) {   
   pq = hh[k] ;
   pplist[k][0] = pq->a1 ;
   pplist[k][1] = pq->a2 ;
/**
   printf("pairh:  %5d %5d %5d %5d %g\n", x, k, pq->a1, pq->a2, pq->score) ;
*/
  }

  for (i=0; i<nlist; i++)  {
    free(hh[i])  ;
  }
  free(hh) ;

  for (i=0; i<l1; i++) {  
   free(wa[i]) ;
  }
  free(wa) ; 

  for (i=0; i<l2; i++) {  
   free(xa[i]) ;
  }
  free(xa) ; 
  fflush(stdout) ;
}

void 
genpair(int **pplist, int nlist, 
  int *ia, int *ib, int n)  

/** 
 strategy 
 divide and conquer 
*/

{
  int i, j, k, l ;
  int **wa, **xa ;
  int va[4], vb[4] ;
  int n1, n2, l1, l2, mul ;

  if (n==0) return ;

  if (n==1)  {  
   genl1(ia[0],ib[0], &l, va, vb) ;
   for (i=0; i<l; i++) {  
    pplist[i][0] = va[i] ;
    pplist[i][1] = vb[i] ;
   }
   return ;
  }

  n1 = n/2 ;
  n2 = n-n1 ;
  l1 = lenplist(ia, ib, n1) ;
  l2 = lenplist(ia+n1, ib+n1, n2) ;

  if ((l1*l2) != nlist) {
   printi1(ia,n) ;
   printi1(ib,n) ;
   fflush(stdout) ;
   fatalx("(genpair) logic bug %d %d %d\n",l1,l2,nlist) ;
  }

  ZALLOC(wa, l1, int *) ;
  for (i=0; i<l1; i++) {  
    ZALLOC(wa[i], 2, int) ;
  }

  ZALLOC(xa, l2, int *) ;
  for (i=0; i<l2; i++) {  
    ZALLOC(xa[i], 2, int) ;
  }

  genpair(wa, l1, ia, ib, n1) ;
  genpair(xa, l2, ia+n1, ib+n1, n2) ;

  mul = (int) pow(2.0, (double) n2) ;
  k=0 ;
  for (i=0; i<l1; i++) {
   for (j=0; j<l2; j++) {
    pplist[k][0] = mul*wa[i][0]+xa[j][0] ;
    pplist[k][1] = mul*wa[i][1]+xa[j][1] ;
    ++k ;
   }
  }
  for (i=0; i<l1; i++) {  
   free(wa[i]) ;
  }
  free(wa) ; 

  for (i=0; i<l2; i++) {  
   free(xa[i]) ;
  }
  free(xa) ; 
}

void
genppp(int **pp, int *xpc, int *xa, int *xb, int l, int ttype)  
{ 
  int k, len, k1, k2 ;
  int **plist ;

  *xpc = 1 ;
  if (l==0) return ;

/**
  printf("genppp %d %d\n",ttype,l) ; 
*/

  switch (ttype)  {
   
  case 0: 
   *xpc = 1; 
   k=0; 
   copyiarr(xa, pp[k], l) ; 
   copyiarr(xb, pp[k]+l, l) ; 
   return ;

  case 1: 
   len = *xpc = pow (2.0, (double) l); 
   ZALLOC(plist, len, int *) ;
   for (k=0; k<len; k++) {  
    ZALLOC(plist[k], l, int) ;
   }
   getpat(plist, l) ;
   for (k=0; k<len; k++) {  
    copyiarr(plist[k], pp[k], l) ;
    copyiarr(plist[len-1-k], pp[k]+l, l) ;
   }
   for (k=0; k<len; k++) {  
    free(plist[k]) ;
   }
   free(plist) ;
   return ;

  case 2: 
   len = *xpc = pow (2.0, (double) l); 
   ZALLOC(plist, len, int *) ;
   for (k=0; k<len; k++) {  
    ZALLOC(plist[k], l, int) ;
   }
   getpat(plist, l) ;
   for (k=0; k<len; k++) {  
    copyiarr(xa, pp[k], l) ;
    copyiarr(plist[len-1-k], pp[k]+l, l) ;
   }
   for (k=0; k<len; k++) {  
    free(plist[k]) ;
   }
   free(plist) ;
   return ;


  case 3: 
   len = *xpc = pow (2.0, (double) l); 
   ZALLOC(plist, len, int *) ;
   for (k=0; k<len; k++) {  
    ZALLOC(plist[k], l, int) ;
   }
   getpat(plist, l) ;
   for (k=0; k<len; k++) {  
    copyiarr(plist[k], pp[k], l) ;
    copyiarr(xb, pp[k]+l, l) ;
   }
   for (k=0; k<len; k++) {  
    free(plist[k]) ;
   }
   free(plist) ;
   return ;


  case 4: 
   len = pow (2.0, (double) l); 
   *xpc = len*len ;
   ZALLOC(plist, len, int *) ;
   for (k=0; k<len; k++) {  
    ZALLOC(plist[k], l, int) ;
   }
   (void) getpat(plist, l) ;
   k=0; 
   for (k1=0; k1<len; k1++) {  
    for (k2=0; k2<len; k2++) {  
     copyiarr(plist[k1], pp[k], l) ;
     copyiarr(plist[k2], pp[k]+l, l) ;
     ++k ;
    }
   }
   for (k=0; k<len; k++) {  
    free(plist[k]) ;
   }
   free(plist) ;
   return ;
 }

}

int getpat (int **plist, int l) 
/** 
 generate all binary patterns length l
*/
{
   int n, k ;
   int *xa ;

   if (l==0) return 1 ;

   if (l==1) {  
    plist[0][0]  = 0;
    plist[1][0]  = 1;  
    return 2 ;
   }


   ZALLOC(xa, l, int) ;
   n = getpat(plist, l-1) ;

   for (k=0; k<n; k++) {  
    if (l>1) copyiarr(plist[k], xa+1, l-1) ;
    xa[0] = 0 ;
    copyiarr(xa, plist[k], l) ;
    xa[0] = 1 ;
    copyiarr(xa, plist[k+n], l) ;
   }

   free(xa);
   return (2*n) ;

}
int
catpat(int **x1, int **x2, int **x3, int nx1, int nx2, 
 int  l1, int l2)  
{
    int k1, k2, k, junk ;
    k = 0 ;
    for (k1=0; k1<nx1; k1++)  {  
     for (k2=0; k2<nx2; k2++)  {  
      if (l1>0) copyiarr(x1[k1], x3[k], l1) ;
      if (l2>0) copyiarr(x2[k2], x3[k]+l1, l2) ;
      if (l1>0) copyiarr(x1[k1]+l1, x3[k]+l1+l2, l1) ;
      if (l2>0) copyiarr(x2[k2]+l2, x3[k]+2*l1+l2, l2) ;
      if (debug) {
       ivmaxmin(x3[k], 2*(l1+l2), NULL, &junk) ;
       if (junk<0) {  
        fprintf (stderr, "yukk %d %d %d  %d %d\n",k1,k2,k,l1,l2) ;
        printi2(x1[k1],2*l1) ;
        printi2(x2[k2],2*l2) ;
        fatalx("yukk\n") ;
       }
      }
      ++k ;
     }
    }
  return k ;
}
void
genl1(int a, int b, int *ll, int *va, int *vb)  
{
/** 
 generate legals for 1 position */

   int t, l ;
   t = -1; 
   if ((a == 9) && (b == 9) && (t<0)) { 
    t=1 ;
    l=2 ;
   }
   if ((a >= 0) && (b >= 0) && (t<0)) { 
    t=0 ;
    l=1 ;
   }
   if ((a < 0) && (b < 0) && (t<0)) { 
    t=4 ;
    l=4 ;
   }
   if ((a >= 0) && (b < 0) && (t<0)) { 
    t=2 ;
    l=2 ;
   }
   if ((a <0)  && (b >= 0) && (t<0)) { 
    t=3 ;
    l=2 ;  
   }
   if (t<0) fatalx("bug (genl1) %d %d\n",a,b) ;

   *ll = l ;
   switch (t)  {   
 
     case 0:  
      va[0] = a ; vb[0] = b ;  return ;

     case 1:  
      va[0] = 0 ; vb[0] = 1 ;  
      va[1] = 1 ; vb[1] = 0 ;  
      return ;

     case 2:  
      va[0] = a ; vb[0] = 0 ;  
      va[1] = a ; vb[1] = 1 ;  
      return ;

     case 3:  
      va[0] = 0 ; vb[0] = b ;  
      va[1] = 1 ; vb[1] = b ;  
      return ;

     case 4:  
      va[0] = 0 ; vb[0] = 0 ;  
      va[1] = 0 ; vb[1] = 1 ;  
      va[2] = 1 ; vb[2] = 0 ;  
      va[3] = 1 ; vb[3] = 1 ;  
      return ;
   }


}

int codeit2(int *ttt, int len)  
{  
  int i, k = 0, t ;

  for (i=0; i<len; i++)  {
   t = ttt[i] ;
   if (t<0) fatalx("badcode\n") ;
   if (t>1) fatalx("badcode\n") ;
   k = k*2+t ;
  }
  return k ;
}

void uncodeit2(int *ttt, int len, int kk)  
{  
  int i, k ;
  k = kk ;
  for (i=0; i<len; i++)  {
   ttt[len-i-1] = k%2 ;
   k /= 2 ;
  }
}

int msk(int n) { 
 if (n>= (8 * sizeof (int)  ) ) fatalx("(msk) n too large %d\n",n) ;
 return ( (1 << n) - 1) ; 
}
int right (int k, int n) 
/* return right n-1 bits */
{
  return (k & msk(n-1)) ;
}
int left (int k, int n) 
/* return left n-1 bits */
{
  return ( (k >>1 ) & msk (n-1))  ;
}
