#include  <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <ranmath.h>

#include <strsubs.h>
#include "heapsubs.h" 

int hash(int khash, int n, int sec);
double zhash(int *a, int n) ;
int hasha(int *a, int n, int modx) ;

void printheap(HEAPX **a, int n, int len) 
{
  int i, k ;
  int *buff ;

  ZALLOC(buff, len, int) ;

  for (i=0; i<n; i++) {
   printf("node: %d %9.3f\n",i, a[i]->score) ;
   uncodeit2(buff, len, a[i]->a1) ;
   for (k=0; k<len; k++) {
    printf("%5d", buff[k]) ;
   }
   printf("\n") ;
   uncodeit2(buff, len, a[i]->a2) ;
   for (k=0; k<len; k++) {
    printf("%5d", buff[k]) ;
   }
   printf("\n") ;
  }
  free(buff) ;
}

void 
add1heap(HEAPX **a , int n, HEAPX *pq)  
{
  if (pq->score < a[0]->score) return ;
  copyheap(pq, a[0]) ;
  siftup(a, 0, n) ;
}

void copyheap(HEAPX *p, HEAPX *q) 
{
  q->a1 = p->a1 ;
  q->a2 = p->a2 ;
  q->score = p->score ; 
}

void 
clearheap(HEAPX **a , int n)  
{

  int i ;
  HEAPX *pt ;
  for (i=0; i<n; i++)  {  
   pt = a[i] ;
   pt->a1 = pt->a2 = 0 ;
   pt->score = -1.0e20 ;
  }

}

void 
heapify(HEAPX **a , int n)  
{
      int l ;
      for (l=n/2; l>=0; --l)  {
       siftup(a,l,n) ;  
      }

}
void 
siftup(HEAPX **a, int l, int n) 
/** 
 the 2 subtrees that are progeny of 
 l are assumed to alread be a heap.  
 We now fix up node l 
*/
{
    HEAPX *temp ;
    int j1, j2, jj ;

    jj= j1 = 2*l+1 ;
    j2 = 2*l+2 ;
    if (j1>=n) return ;
    if ((j2<n) && (a[j2]->score<a[j1]->score)) jj = j2 ;  
/* jj is now smallest child of l */
    if (a[l]->score <= a[jj]->score) return ;  /* nothing to do */
    temp = a[l];  a[l]=a[jj] ;  a[jj]=temp ;
    siftup(a,jj,n) ;
}
int hasha(int *a, int n, int modx) 
{

   int msk ;
   int p1, mul, k, t ;
   int i ;

   t = sizeof(int) * 4 ;
   msk = (1<<(t-1)) -1 ; 

   p1 = 524287 ;  /* 2^19 -1 ) prime */
   mul = 2011 ;
   k=1; 
   for (i=0; i<n; i++) {  
    t = abs(a[i]) % p1 ;  
    k = (k << 7) + t*mul ;  
    k &= msk ;  
   }
   return hash(k, modx, 1)  ;
}

double zhash(int *a, int n) 

{
  int x ;
  double y ;
  int modx = 100000 ;
  x = hasha(a,n, modx) ; 
  y = (double) x / (double) modx ;
  return y ;
}

int hash(int khash, int n, int sec)
{
  int x1 ;
  int x2 ;
  int t, jhash ;
  int a1, a2, a3 ;

/**
   initial hash to messup
   structure such as arith progressions
*/
  a1 = khash % ((1<<23) - 1) ;
  t = (khash << (sec+5)) * 97  ;
  a2 = t % ( (1<<21) - 1) ;

  jhash = abs(a1 ^ a2) ;

  x1 = jhash%n ;
  t = (khash%20011) * 101;
  x2 = t%n ;

  t = (x1+x2*sec)%n ;
  if (t==0) t=1 ;
  return  t ;

}

int idedupe (int *x, int m, int n) 
// probably better in sortit.c 
// takes m*n integer array and sorts and dedupes rows 
// trivial to modify to just sort
{
  int **xpt, *indx, *y ;
  int i, kout ;

  if ((m==0) || (n==0)) return 0 ;

  ZALLOC(xpt, m, int *) ;
  ZALLOC(indx, m, int) ;
  ZALLOC(y, m*n, int) ;
  for (i=0; i<m; i++) {  
   xpt[i] = x+i*n ;
  }
  ipsortit(xpt, indx, m, n) ;
  copyiarr(xpt[0], y, n) ;
  kout = 1 ;
  for (i=1; i<m; i++) {  
   if (compiarr(xpt[i], xpt[i-1], n) == 0) continue ;
   copyiarr(xpt[i], y+kout*n, n) ;
   ++kout ;
  }
  copyiarr(y, x, kout*n) ;
  free(xpt) ;
  free(indx) ;
  free(y) ;
  return kout ;
}
