#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <nicklib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "getpars.h"

#define  VERSION  "1.52"
#define  CNULL '\0'  

double C= 23, G=33, lambdal = 4, lambdaa=6, r1 = 2, r2 = -1 ;
double theta = 0.2 ;
double p=0.2, pp=0.8 ;
double nassoc = 1.0e6  ;  
double fpos = .05 ;
double fneg = .20 ;
double dfreq = 0.0 ; /* disease prob */
int usepp = NO ;
int useadmix = NO ;
int twosided = NO ;
int verbose = NO ;

double assocsamp, linksamp, ancestsamp ;

void readcommands(int argc, char **argv) ;

double 
doassoc(double pp, double  r1, double r2, 
 double nassoc, double fpos, double fneg)  ;

double 
dolinkage(double pp, double  r1, double r2, 
 double lambda, double fpos, double fneg)   ;

double 
doancest(double p, double pp, double theta, 
 double  r1, double r2, 
 double lambda, double fpos, double fneg)  ;

void
mktmat(double *tmat, double *xa, double *xb)   ;

void
 mkx(double *x, double *a, double *b) ;

double
nxcalc(char *pstr, double rmean, double rsig, double cmean, double csig, 
 double fpos, double fneg)  ;

double
klcalc(char *pstr, double lambda, double cmean, double csig, 
 double fpos, double fneg)  ;

double klsig(double C, double G, double lambda, double T)  ;
double klsiginv(double C, double G, double lambda, double mu) ;
double knewt(double C, double G, double lambda, double z, double ptail);

void sett1(double *tt, double theta, int numstates) ;
void pcheck (char *name, char x)  ;

int main(int argc, char **argv)
{
 double xp ;

 assocsamp = linksamp = ancestsamp ;
 readcommands( argc, argv)  ;
 if (r2 <0 ) r2 = r1*r1 ;

 if (verbose) {
  printf("risk(1): %9.3f\n",r1) ;
  printf("risk(2): %9.3f\n",r2) ;
  printf("\n\n") ;
 }

 xp = p ;

 if (useadmix) {
  usepp = NO ;
  xp = theta * pp + (1.0-theta) * p ; 
  if (verbose) 
  printf("assoc/linkage done using admix allele frequency\n") ;
 }

 if (usepp) {
  xp = pp ; 
  if (verbose) 
  printf("assoc/linkage done using pp allele frequency\n") ;
 }

 if ((xp ==0.0) || (xp ==1.0)) {
  if (verbose)
  printf ("no association or linkage analysis -- mono marker\n") ;
 }
 else { 
  assocsamp = doassoc(xp, r1, r2, nassoc, fpos, fneg) ;
  if (verbose)
  printf("\n\n") ;
  linksamp =  dolinkage(xp, r1, r2, lambdal, fpos, fneg) ;
  if (verbose) 
  printf("\n\n") ;
 }
 ancestsamp = doancest(p, pp, theta, r1, r2, lambdaa, fpos, fneg) ;
 if (!verbose) printf("%9.3f %9.3f %9.3f\n", 
  assocsamp, linksamp, ancestsamp) ;
}
double 
doassoc(double xp, double  r1, double r2, 
 double nassoc, double fpos, double fneg)  
{
   double tt[3], ttr[3], tth[3], y, uu[3], u0  ;

/** 
 tth is allele freq in healthies 
*/

   double rmean, rvar, rsig, rtail , rz, rzz ;
   double cmean, cvar, csig, ctail , cz, mu ;
   double sqrtn, ssize ;
   double rmean2, rvar2, cmean2, cvar2, rsig2, csig2 ;
   double ffpos ;
   double h1 , h2 ;
/** for case-control differences */
   sett1(tt, xp, 3) ;
   ttr[0] = tt[0] ;
   ttr[1] = tt[1] * r1 ;
   ttr[2] = tt[2] * r2 ;
   y = asum(ttr, 3) ; 
   u0 = dfreq/y ;
/* 
 u0 = P(D| a = 0) 
*/

   if (u0>=1.0) fatalx("dfreq too large!\n") ;
   vst(ttr, ttr, 1.0/y, 3) ;

   h1 = (1.0-u0*r1) / (1.0-u0) ;
   h2 = (1.0-u0*r2) / (1.0-u0) ;

   if (h1 < 0.0) fatalx("dfreq too large!\n") ;
   if (h2 < 0.0) fatalx("dfreq too large!\n") ;

   tth[0] = tt[0] ;
   tth[1] = tt[1] * h1 ;
   tth[2] = tt[2] * h2 ;
   y = asum(tth, 3) ; 
   vst(tth, tth, 1.0/y, 3) ;

   copyarr(tth, tt, 3) ;
   
   rmean = tt[1] + 2 *tt[2] ;
   rvar = tt[1] + 4 * tt[2] - rmean*rmean ;
   rsig = sqrt(rvar) ;
   cmean = ttr[1] + 2 *ttr[2] ;
   cvar = ttr[1] + 4 * ttr[2] - cmean*cmean ;
   csig = sqrt(cvar) ;
   rmean2 = 0.0 ;
   rsig2 = sqrt(2.0) * rsig ;
   cmean2 = cmean - rmean ;
   cvar2  = rvar + cvar ;
   csig2 = sqrt(cvar2) ;
   if (verbose) {
    printf("assoc. analysis\n") ;
    if (twosided) 
     printf("twosided!:\n") ;
    printf("p(control): %9.3f p(case): %9.3f\n", rmean/2, cmean/2) ;
    if (dfreq>0) 
     printf("disease frequency: %12.6f\n", dfreq) ;
    printf("number of SNPs tested: %9.3f\n",nassoc) ;
   }
    ffpos = fpos/nassoc ;
    if (twosided) ffpos /= 2.0 ;
    return 
     nxcalc("case-control pairs",
        rmean2, rsig2, cmean2, csig2, ffpos, fneg) ;
}

double
nxcalc(char *pstr, double rmean, double rsig, double cmean, double csig, 
 double fpos, double fneg)  
{

   double rtail , rz, rzz ;
   double cz, mu ;
   double sqrtn, ssize, fom ;
   double xrmean, xrsig, xcmean, xcsig ;

   rtail = fpos ;  
   rz = zprob(rtail) ;  
   cz = zprob(1-fneg) ;  /* sigmage for causal score */
   fom = mu = xcmean = (cmean - rmean)/ rsig ;
   xcsig = csig/rsig ;
   sqrtn = (rz - xcsig*cz ) / mu  ;
   ssize =  sqrtn * sqrtn ;
   if (verbose) {
    printf("figure of merit: %9.3f\n", xcmean) ;
    printf("rz: %9.3f\n", rz) ;
    printf("cz: %9.3f\n", cz) ;
    printf("%s: %9.3f\n", pstr, ssize) ;
   }
   return ssize ;
}

double
klcalc(char *pstr, double lambda, double cmean, double csig, 
 double fpos, double fneg)  

{

   double rtail , rz, rzz, rmean = 0.0, rsig = 1.0 ;
   double cz, mu, klmu, xcsig ;
   double sqrtn, ssize, fom ;

   rtail = fpos ;  
   klmu = -log(1-fpos) ; 
/** 
 probability we get hit 
 if poisson rate of KL clumps is klmu 
 shall be fpos 
*/
   rz = klsiginv(C, G, lambda, klmu) ;
   cz = zprob(1-fneg) ;  /* sigmage for causal score */
   fom = mu =  (cmean - rmean)/ rsig ;
   xcsig = csig/rsig ;
   sqrtn = (rz - xcsig*cz ) / mu  ;
   ssize =  sqrtn * sqrtn ;
   if (verbose) {
   printf("figure of merit: %9.3f\n", cmean) ;
   printf("rz: %9.3f\n", rz) ;
   printf("cz: %9.3f\n", cz) ;
   printf("%s: %9.3f\n", pstr, ssize) ;
   }
   return ssize ;
}


double 
dolinkage(double pp, double  r1, double r2, 
 double lambda, double fpos, double fneg)  

{
   double tt[3], ttr[3], y, qq ;
   double tibd[4], tcibd[4], tmat[9] ;
   double pdd[3], risk[3], pr[3], pc[3] ;
   double mur, vr, muc, vc, fom, g, csig, rsig;
   int l1, l2, x ;

   risk[0] = 1 ;
   risk[1] = r1 ;
   risk[2] = r2 ;

   pr[0] = 0.25 ;
   pr[1] = 0.50 ;
   pr[2] = 0.25 ;

   sett1(tt, pp, 3) ;
   qq = 1-pp ;

   vzero(tibd, 4) ;
   vzero(tcibd, 4) ;

   tibd[0] = qq ;
   tibd[3] = pp ;

   tcibd[0] = qq*qq ;
   tcibd[3] = pp*pp ;
   tcibd[1] = tcibd[2] = pp*qq ;

/* prob of allele pairs given IBD or not IBD */
   for (x=0; x<= 2; ++x)  {   
   /* x is number of IBD */
    if (x==0) mktmat(tmat, tcibd, tcibd) ;
    if (x==1) mktmat(tmat, tibd, tcibd) ;
    if (x==2) mktmat(tmat, tibd, tibd) ;
    
    for (l1=0; l1 <=2 ; ++l1) 
     for (l2=0; l2 <=2 ; ++l2) {
      tmat[3*l1+l2] *= risk[l1]*risk[l2] ;
    }
    pdd[x] = asum(tmat,9) ;
  }
  vvt(pc, pr, pdd,3) ;
  y = asum(pc, 3) ;
  vst(pc, pc, 1.0/y, 3) ;
  mur = 1 ;
  vr = 0.5 ;
  muc = pc[1] + 2*pc[2] ;
  vc = pc[1] + 4*pc[2] - muc*muc ;

  rsig = sqrt(vr) ;
  csig = sqrt(vc) ;

  fom = (muc-mur)/rsig ;
/* figure of merit */
   if (verbose) {
   printf("linkage analysis\n") ;
   printf("p(IBD): %9.3f\n", muc/2) ;
   }

   return klcalc("sib-pairs", lambda, fom, csig/rsig, fpos, fneg) ;

}

double 
doancest(double p, double pp, double theta, 
 double  r1, double r2, 
 double lambda, double fpos, double fneg)  

{
   double tt[3], ttr[3], y, qq ;
   double ta[2], tb[2] ;
   double tmat[3] ;
   double pdd[3], risk[3], pr[3], pc[3] ;
   double mur, vr, muc, vc, fom, g, csig, rsig;
   int l1, l2, x ;

   risk[0] = 1 ;
   risk[1] = r1 ;
   risk[2] = r2 ;

   sett1(pr, theta, 3) ;

   ta[0] = 1-p ;
   ta[1] = p ;
/* prob of allele given e=0 */

   tb[0] = 1-pp ;
   tb[1] = pp ;
/* prob of allele given e=1 */

   for (x=0; x<= 2; ++x)  {   
/* x is number of C-alleles */
    if (x==0) mkx(tmat, ta, ta) ;       
    if (x==1) mkx(tmat, ta, tb) ;       
    if (x==2) mkx(tmat, tb, tb) ;       
    
    vvt(tmat, tmat, risk, 3) ;
    pdd[x] = asum(tmat,3) ;
  }
  vvt(pc, pr, pdd,3) ;
  y = asum(pc, 3) ;
  vst(pc, pc, 1.0/y, 3) ;

  y = pdd[0] ;
  vst(pdd, pdd, 1.0/y, 3) ;

  mur = pr[1] + 2*pr[2] ;
  vr = pr[1] + 4*pr[2] - mur*mur ;

  muc = pc[1] + 2*pc[2] ;
  vc = pc[1] + 4*pc[2] - muc*muc ;

  rsig = sqrt(vr) ;
  csig = sqrt(vc) ;

  fom = (muc-mur)/rsig ;
/* figure of merit */
   if (verbose) {
    printf("ancestry analysis\n") ;
    printf("theta(control): %9.3f\n", mur/2) ;
    printf("theta(case): %9.3f\n", muc/2) ;
    printf("erisk(1): %9.3f\n", pdd[1]) ;
    printf("erisk(2): %9.3f\n", pdd[2]) ;
   }

   return klcalc("admixed cases", lambda, fom, csig/rsig, fpos, fneg) ;

}
void
mktmat(double *tmat, double *xa, double *xb)  
{
  int a1, b1, a2, b2, a3, b3 ;

  vzero(tmat, 9) ;
  for (a1=0; a1<=1; ++a1) 
   for (a2=0; a2<=1; ++a2) 
    for (b1=0; b1<=1; ++b1) 
     for (b2=0; b2<=1; ++b2)  {
       a3 = a1+a2 ;
       b3 = b1+b2 ;
       tmat[3*a3+b3] += xa[2*a1+b1]*xb[2*a2+b2] ;
     }
}


void readcommands(int argc, char **argv) 

{
  char *parname = NULL ;
  phandle *ph ;
  char c;

  while ((c = getopt (argc, argv, "p:v")) != -1) {

    switch (c)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
        printf("version: %s\n", VERSION) ;
	parname = strdup(optarg) ;
	break;

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   ph = openpars(parname) ;

   getdbl(ph,"C:", &C) ;
   getdbl(ph,"G:", &C) ;
   getdbl(ph,"lambdal:", &lambdal) ;
   getdbl(ph,"lambdaa:", &lambdaa) ;
   getdbl(ph,"theta:", &theta) ;
   getdbl(ph,"p:", &p) ;
   getdbl(ph,"pp:", &pp) ;
   getdbl(ph,"r1:", &r1) ;
   getdbl(ph,"r2:", &r2) ;
   getdbl(ph,"fpos:", &fpos) ;
   getdbl(ph,"fneg:", &fneg) ;
   getdbl(ph,"dfreq:", &dfreq) ;
   getdbl(ph,"nassoc:", &nassoc) ;
   getint(ph,"usepp:", &usepp) ;
   getint(ph,"useadmix:", &useadmix) ;
   getint(ph,"twosided:", &twosided) ;
   getint(ph,"verbose:", &verbose) ;
}

double klsig(double C, double G, double lambda, double T) 
{
   double alpha, mu  ;
   alpha = ntail(T) ;
   mu = (C + lambda*G*T*T)*alpha ;
   return mu ;
}

double klsiginv(double C, double G, double lambda, double mu) 
/** 
 return thresghold T so that random KL prob is mu
*/
{
  double t, tx ;
  if (mu>0.5) 
   return 0.0 ;
  if (mu <= 0.0) return 50 ;
  tx = 0.0 ;  
  for (;;) {  
   ++tx ;
   if (klsig(C,G, lambda, tx) < mu) { 
    tx -= 0.5 ; break ;
   }
  }
  t = knewt(C, G, lambda, tx, mu) ;
  return t ;
}
double knewt(double C, double G, double lambda, double z, double ptail)
{
/**
 newton step
 z is approximation
*/
    double p0, pder, h ;
    double pi, zz ;
    double a, b, ader, bder, xa, xb, ltail ;
    int iter ;

    pi = 2.0*acos(0.0) ;
    zz = z ;
    ltail = log(ptail) ;
    
    for (iter = 1; iter<=10; ++iter) {
     p0 = klsig(C, G, lambda, zz) ; p0 = log(p0) ;
/**
     printf("%g %g %g %20.10f\n", zz, p0, ltail, ltail-p0) ;
*/
     xa = ntail(zz) ; a = log(xa) ;
     xb = C + lambda*G*zz*zz ; b = log(xb) ;
     ader = -exp(-0.5*zz*zz)/(sqrt(2*pi)*xa) ;
     bder = 2*lambda*G*zz/xb ;
     pder = bder + ader ;
     if (pder==0.0) return zz ;
     h = (ltail-p0)/pder ;
     if (h==0.0) return zz ;
     zz += h ;
    }
    return zz ;
}
void
 mkx(double *x, double *a, double *b)        
{

  x[0] = a[0]*b[0] ;
  x[1] = a[1]*b[0] + a[0]*b[1] ;
  x[2] = a[1]*b[1] ;

}

void sett1(double *tt, double theta, int numstates) 
{
    if (numstates==2)  { 
     tt[0] = 1.0-theta ;
     tt[1] = theta ;
     tt[2] = 0.0 ;
    }
    if (numstates==3)  { 
     tt[0] = (1.0-theta)*(1.0-theta) ;
     tt[1] = 2.0*theta*(1.0-theta) ;
     tt[2] = theta*theta ;
    }
}

void pcheck (char *name, char x) 
{
 
 if (name == NULL) 
  if (x != CNULL) 
   fatalx ("parameter %c compulsory\n",x) ;
  else fatalx("missing argument\n") ;
}

