#include <stdio.h>
#include <limits.h>
#include <math.h>  
#include <nicklib.h> 
#include <getpars.h> 
#include "admutils.h"  

/* ********************************************************************* */

int verbose = NO ;

int mval = -1 ;
int nval = -1 ;
int numsamp = 100;
long seed= 0  ;

char *outputname = NULL ;


int estxx(double *lam, int m) ;  
double xxlike(int m, double a, double var, double logsum , double lsum)    ;
double xxlikex(int m, double a, double logsum , double lsum)    ;
double xxliked(int m, double a, double logsum , double lsum)    ;
double xxliked2(int m, double a, double logsum , double lsum)    ;
void readcommands(int argc, char **argv) ;
int readit(double *xx) ;

double zn, zvar  ;

#define NMAX 10000  

int main(int argc, char **argv)
{
    int numiter = 1 ;
    int iter, m, n, k, i;
    double *a, *b, *lam ;
    double y1, y2, y ;
    int numtest = 100000 ;
    double tw1, tw2, tail1, tail2 ;    
    FILE *ofile ;
    int *cls,  *ind, n0, n1, clen ;
    double kstail, *stats ;


/**
    readcommands(argc, argv) ;
*/

    ZALLOC(a, NMAX, double) ;
    m = readit(a) ;
    vst(a, a, -1.0, m) ;
    sortit(a, NULL, m) ;
    vst(a, a, -1.0, m) ;

    for (i=0; i<m; i++) { 
      printf("%4d %9.3f\n", i, a[i]) ;
    }

     for (k=0; k<MIN(m/2, 20); ++k) {
      estxx(a+k, m-k) ;   
      y = a[k]/zvar ;
      tw2 = twnorm(y, (double) (m-k), zn) ;
      tail2 = twtail(tw2) ;
      printf("%4d %12.3f  %9.3f %9.3f     %9.3f %9.3f\n", k, zn, a[k], y, tw2, tail2) ;
     }
}

int estxx(double *lam, int m) 
{
  double lsum, logsum ;
  double *ww ;
  double  a, p, yn, var ;
  double ylike, ybase, y, ylmax, ynmax, yld, yld2, ainc, ym ;
  int k ;

/**
  *pn = 99.0 ;
  *pvar = 88.0 ;
  return 1 ;
*/

  ZALLOC(ww, m, double) ;
  copyarr(lam, ww, m) ;
  lsum = asum(ww, m) ;
  vlog(ww, ww, m) ;
  logsum = asum(ww, m) ;
  
  ylmax = -1.0e20 ;
  yn = (double) m ;
  for (k= 1; k<=100; ++k) {  
   a = yn/2.0 ;
   ylike = xxlikex(m, a, logsum, lsum) ; 
   yld = xxliked(m, a, logsum, lsum) ; 
   ylike -= ybase ;
   if (verbose) 
    printf("ynloop %12.3f %12.3f %12.3f\n", yn / (double) m , ylike, yld) ;
   if (ylike < ylmax) break ;
   ylmax = ylike ;
   ynmax = yn ;
   yn *= 1.1 ;
  }
  a = ynmax/2.0 ;
  for (k= 1; k<=10; ++k) {  
// newton iteration
   ylike = xxlikex(m, a, logsum, lsum) ; 
   yld  = xxliked(m, a, logsum, lsum) ; 
   yld2 = xxliked2(m, a, logsum, lsum) ; 
   ylike -= ybase ;
   ainc = -yld/yld2 ;
   a += ainc ;
   if (verbose) 
   printf("newton: %3d  %15.9f  %15.9f  %15.9f\n", k, ylike, yld, ainc)  ;
  }
  fflush(stdout) ;
  yn = 2.0*a ;
  ym = (double) m ;
  var = lsum/ (2.0*a*ym) ;

  zn = yn ;
  zvar = var ;
  free(ww) ;
  return 0 ;
}
double xxlike(int m, double a, double var, double logsum , double lsum)   
{
  double p , yl = 0.0 ;
  double ym , x ;
  int j ;

  p = 0.5* (double) (m+1) ;
  ym = (double) m ;

  yl = -ym*a*log(2.0) ;
  for (j=1;  j<= m; ++j) {  
   x = a - 0.5* (double) (m-j) ;
   yl -= lgamma(x) ;
  }
// so far this is log (C_L)    normalizing constant
  yl -= ym*a*log(var) ;
  yl += (a-p)*logsum ;
  yl -= lsum/(2.0*var) ;

  return yl ;

}
double xxlikex(int m, double a, double logsum , double lsum)   
{
  double p , yl = 0.0 ;
  double ym , x, var, lco ;
  int j ;

  p = 0.5* (double) (m+1) ;
  ym = (double) m ;
  lco = lsum/(2.0*ym) ;
  var = lco/a ;


  yl = -ym*a*log(2.0) ;
  for (j=1;  j<= m; ++j) {  
   x = a - 0.5* (double) (m-j) ;
   yl -= lgamma(x) ;
  }
// so far this is log (C_L)    normalizing constant
  yl -= ym*a*log(var) ;
  yl += (a-p)*logsum ;
  yl -= ym*a ;  // plugging in var 

  return yl ;

}
double xxliked(int m, double a, double logsum , double lsum)   
// first deriv wrt a
{
  double p , yl = 0.0 ;
  double ym , x, var, vard ;
  int j ;

  p = 0.5* (double) (m+1) ;
  ym = (double) m ;
  var = lsum/ (2.0*a*ym) ;
  vard = -var/a ;


  yl = -ym*log(2.0) ;
  for (j=1;  j<= m; ++j) {  
   x = a - 0.5* (double) (m-j) ;
   yl -= psi(x) ;
  }
// so far this is log (C_L)    normalizing constant
  yl -= ym*log(var) ;
  yl -= (ym*a/var)*vard ;
  yl += logsum ;
  yl -= ym ;  // plugging in var 

  return yl ;

}
double xxliked2(int m, double a, double logsum , double lsum)   
// second deriv wrt a
{
  double p , yl = 0.0 ;
  double ym , x, var, vard, vard2, y ;
  int j ;

  p = 0.5* (double) (m+1) ;
  ym = (double) m ;
  var = lsum/ (2.0*a*ym) ;
  vard = -var/a ;
  vard2 = 2.0*var/(a*a) ;


  yl = 0.0 ; 
  for (j=1;  j<= m; ++j) {  
   x = a - 0.5* (double) (m-j) ;
   yl -= tau(x) ;
  }
// so far this is log (C_L)    normalizing constant
  yl -= 2.0*(ym/var)*vard ;
  yl -= (ym*a/var)*vard2 ;
  y = vard/var ;
  yl += (ym*a)*y*y ;

  return yl ;

}
   
void readcommands(int argc, char **argv) 

{
  int i ;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:V")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

  if (parname==NULL) fatalx("-p compulsory\n") ;
   ph = openpars(parname) ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;

   getstring(ph, "output:", &outputname) ;
   getint(ph, "mval:", &mval) ;
   getint(ph, "nval:", &nval) ;
   getint(ph, "numsamp:", &numsamp) ;
   n = 0 ;
   getint(ph, "seed:", &n) ;
   seed = n  ;

   printf("### THE INPUT PARAMETERS\n");
   printf("##PARAMETER NAME: VALUE\n");
   writepars(ph);
   closepars(ph) ;

}

int 
readit(double *xx)
{
#define MAXSTR 128
#define MAXFF 10

  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, j, num=0 ;

  while (fgets(line, MAXSTR, stdin) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') {
    freeup(spt, nsplit) ;
    continue ;
   }
   if (nsplit<1) { 
     continue ;
   }
   if (num>=NMAX) fatalx("too much data\n") ;
   xx[num] = atof(spt[0]) ;
   freeup(spt, nsplit) ;
   ++num ;
  }
  return num ;
}

